#ifndef LSM_STORE
#define LSM_STORE

#include <stdbool.h>
#include <stdint.h>

#include "lsm.h"
#include "lsm/str.h"

#define LSM_STORE_DATA_LEVELS 3

/**
 * Read-only handle to an entry in the store
 */
typedef struct lsm_read_handle lsm_read_handle;

/**
 * Writeable handle to an entry in the store
 */
typedef struct lsm_write_handle lsm_write_handle;

/**
 * A store consisting of LSM entries.
 *
 * A store manages both an in-memory data structure for quick lookup, and a
 * database file for persistent storage of the contained entries.
 */
typedef struct lsm_store lsm_store;

/**
 * Allocate and initialize a new lsm_store object.
 *
 * @param ptr where to store newly allocated pointer
 */
lsm_error lsm_store_init(lsm_store **ptr);

/**
 * Return how many elements are stored in the trie.
 *
 * @param store store to use
 * @return how many elements are in the store
 */
uint64_t lsm_store_size(const lsm_store *store);

/**
 * Open the given database file and load it into a new store object.
 *
 * @param ptr pointer to store newly allocated store
 * @param data_path path to the data directory
 */
lsm_error lsm_store_load(lsm_store **ptr, lsm_str *data_path);

/**
 * Dealocate an existing lsm_store object.
 *
 * @param store object to deallocate
 */
void lsm_store_free(lsm_store *store);

/**
 * Open a read handle to the given entry. This entry must be properly closed
 * using `lsm_store_handle_close`.
 *
 * @param out pointer to store handle pointer
 * @param store store to retrieve entry from
 * @param key key to search
 */
lsm_error lsm_store_open_read(lsm_read_handle **out, lsm_store *store,
                              const lsm_str *key);

/**
 * Open a write handle to the given entry. This handle should only be used for
 * writing; read operations on this handle are unsupported. This entry must be
 * properly closed using `lsm_store_handle_close`.
 *
 * @param out pointer to store handle pointer
 * @param store store to retrieve entry from
 * @param key key to search
 */
lsm_error lsm_store_open_write(lsm_write_handle **out, lsm_store *store,
                               const lsm_str *key);

/**
 * Insert a new entry into the store, returning a write handle to the newly
 * created entry.
 *
 * @param out pointer to store new entry pointer in
 * @param store store to modify
 * @param key key to add; ownership of key pointer is taken over
 */
lsm_error lsm_store_open_new(lsm_write_handle **out, lsm_store *store,
                             lsm_str *key);

bool lsm_read_attr_present(lsm_read_handle *handle, uint8_t type);
lsm_error lsm_read_attr_get(const lsm_str **out, const lsm_read_handle *handle,
                            uint8_t type);
lsm_error lsm_read_attr_get_uint64_t(uint64_t *out,
                                     const lsm_read_handle *handle,
                                     uint8_t type);
lsm_error lsm_read_attr_get_uint8_t(uint8_t *out, const lsm_read_handle *handle,
                                    uint8_t type);
uint64_t lsm_read_data_len(const lsm_read_handle *handle);
lsm_error lsm_read_data_read(uint64_t *out, char *buf, lsm_read_handle *handle,
                             uint64_t len);
void lsm_read_close(lsm_read_handle *handle);

bool lsm_write_attr_present(const lsm_write_handle *handle, uint8_t type);
lsm_error lsm_write_attr_get(const lsm_str **out,
                             const lsm_write_handle *handle, uint8_t type);
lsm_error lsm_write_attr_get_uint64_t(uint64_t *out,
                                      const lsm_write_handle *handle,
                                      uint8_t type);
lsm_error lsm_write_attr_get_uint8_t(uint8_t *out,
                                     const lsm_write_handle *handle,
                                     uint8_t type);
lsm_error lsm_write_attr_remove(lsm_str **out, lsm_write_handle *handle,
                                uint8_t type);
lsm_error lsm_write_attr_insert(lsm_write_handle *handle, uint8_t type,
                                lsm_str *data);
lsm_error lsm_write_attr_insert_uint64_t(lsm_write_handle *handle, uint8_t type,
                                         uint64_t data);
lsm_error lsm_write_attr_insert_uint8_t(lsm_write_handle *handle, uint8_t type,
                                        uint8_t data);
uint64_t lsm_write_data_len(const lsm_write_handle *handle);
lsm_error lsm_write_data_append(lsm_write_handle *handle, const lsm_str *data);
void lsm_write_remove(lsm_write_handle *handle);
void lsm_write_close(lsm_write_handle *handle);
lsm_error lsm_write_commit(lsm_write_handle *handle);

#endif
