#ifndef LSM_STORE_INTERNAL
#define LSM_STORE_INTERNAL

#include <pthread.h>
#include <stdio.h>

#include "lsm/store.h"
#include "lsm/str_internal.h"
#include "lsm/trie.h"

#define LSM_DB_FILE_NAME "lsm.db"
#define LSM_IDX_FILE_NAME "lsm.idx"
#define LSM_DATA_FILE_SUFFIX ".data"

typedef struct lsm_attr {
  uint8_t type;
  lsm_str *str;
} lsm_attr;

/**
 * An entry inside an LSM store.
 *
 * Each entry consists of the key it's stored behind, zero or more attributes
 * (metadata) and a data file.
 */
typedef struct lsm_entry {
  lsm_str *key;
  struct {
    uint64_t bitmap[4];
    uint8_t count;
    lsm_attr *items;
  } attrs;
  uint64_t data_len;
  uint64_t idx_file_offset;
} lsm_entry;

/**
 * Allocate and initialize a new lsm_entry object.
 *
 * @param ptr where to store newly allocated pointer
 */
lsm_error lsm_entry_init(lsm_entry **ptr);

/**
 * Deallocate an existing entry
 *
 * @param entry pointer to entry
 */
void lsm_entry_free(lsm_entry *entry);

/**
 * Allocates a new entry that's identical to the provided one.
 */
lsm_error lsm_entry_clone(lsm_entry **out, const lsm_entry *entry);

/**
 * Deallocate an existing lsm_entry object.
 *
 * @param entry object to deallocate
 */
void lsm_entry_free(lsm_entry *entry);

typedef struct lsm_entry_wrapper {
  pthread_rwlock_t lock;
  lsm_entry *entry;
} lsm_entry_wrapper;

lsm_error lsm_entry_wrapper_init(lsm_entry_wrapper **ptr);
void lsm_entry_wrapper_free(lsm_entry_wrapper *wrapper);

typedef enum lsm_entry_handle_state : uint8_t {
  lsm_entry_handle_state_new = 1 << 0,
  lsm_entry_handle_state_updated = 1 << 1,
  lsm_entry_handle_state_removed = 1 << 2,
} lsm_entry_handle_state;

struct lsm_store {
  lsm_trie *trie;
  lsm_str *data_path;

  struct {
    FILE *f;
    uint64_t size;
    pthread_mutex_t lock;
  } db;

  struct {
    FILE *f;
    uint64_t size;
    uint64_t block_count;
    pthread_mutex_t lock;
  } idx;
};

/**
 * Read in the database and construct the in-memory trie index. This function
 * assumes the provided store is a newly initialized empty store with the
 * database files opened.
 *
 * @param store store to read
 */
lsm_error lsm_store_load_db(lsm_store *store);

/**
 * Write a new insert to the database.
 *
 * @param handle handle to added entry
 */
lsm_error lsm_entry_disk_insert(lsm_store *store, lsm_entry *entry);

/**
 * Remove an entry from the database.
 *
 * @param handle handle to the removed entry
 */
lsm_error lsm_entry_disk_remove(lsm_store *store, lsm_entry *entry);

/**
 * Update an existing entry already in the store.
 *
 * @param handle to updated entry
 */
lsm_error lsm_entry_disk_update(lsm_store *store, lsm_entry *entry);

/**
 * Return the length of the path to this entry's data file
 */
uint64_t lsm_entry_data_path_len(const lsm_store *store,
                                 const lsm_entry *entry);

/**
 * Fill in the entry's data file path in the provided buffer. Use
 * `lsm_entry_data_path_len` to allocate an appropriately-sized buffer
 */
void lsm_entry_data_path(char *path, const lsm_store *store,
                         const lsm_entry *entry);

/**
 * Open the entry's data file for reading
 *
 * @param handle handle to the entry
 */
lsm_error lsm_entry_data_open_read(lsm_read_handle *handle);

lsm_error lsm_entry_data_open(FILE **out, const lsm_store *store,
                              const lsm_entry *entry, const char *mode);
lsm_error lsm_entry_data_mkdirs(const lsm_store *store, const lsm_entry *entry);

/**
 * Remove the entry's data file if present and close its handle. Any uncommitted
 * changes will be reverted.
 *
 * @param handle handle to the entry
 */
lsm_error lsm_entry_data_remove(const lsm_store *store, const lsm_entry *entry);

/**
 * Checks whether the entry has an attribute with the specified type.
 *
 * @param entry entry to check
 * @param type type of attribute to check for
 */
bool lsm_entry_attr_present(const lsm_entry *entry, uint8_t type);

/**
 * Retrieve the contents of an attribute from an entry, if present
 *
 * @param out where to store pointer to attribute data
 * @param entry entry to search for
 * @param type type of attribute to return
 */
lsm_error lsm_entry_attr_get(const lsm_str **out, const lsm_entry *entry,
                             uint8_t type);

/**
 * Convenience wrapper around `lsm_entry_attr_get` that can be used if we know
 * beforehand the attribute value is a 64-bit number.
 *
 * @param out where to store attribute data
 * @param entry entry to search for
 * @param type type of attribute to return
 */
lsm_error lsm_entry_attr_get_uint64_t(uint64_t *out, const lsm_entry *entry,
                                      uint8_t type);

/**
 * Convenience wrapper around `lsm_entry_attr_get` that can be used if we know
 * beforehand the attribute value is an 8-bit number.
 *
 * @param out where to store attribute data
 * @param entry entry to search for
 * @param type type of attribute to return
 */
lsm_error lsm_entry_attr_get_uint8_t(uint8_t *out, const lsm_entry *entry,
                                     uint8_t type);

/**
 * Add a new attribute to the entry. This overwrites an existing version of this
 * attribute.
 *
 * @param entry entry to modify
 * @param type type of attribute to add
 * @param data data of attribute; ownership of pointer is taken over
 */
lsm_error lsm_entry_attr_insert(lsm_entry *entry, uint8_t type, lsm_str *data);

/**
 * Convenience wrapper around `lsm_entry_attr_insert` that can be used if the
 * data to be stored is a 64-bit number.
 *
 * @param entry entry to modify
 * @param type type of attribute to add
 * @param data data of attribute
 */
lsm_error lsm_entry_attr_insert_uint64_t(lsm_entry *entry, uint8_t type,
                                         uint64_t data);

/**
 * Convenience wrapper around `lsm_entry_attr_insert` that can be used if the
 * data to be stored is an 8-bit number.
 *
 * @param entry entry to modify
 * @param type type of attribute to add
 * @param data data of attribute
 */
lsm_error lsm_entry_attr_insert_uint8_t(lsm_entry *entry, uint8_t type,
                                        uint8_t data);

/**
 * Remove an atribute from the given entry, if present.
 *
 * @param out pointer to store removed data pointer in. If NULL, data pointer
 * will be leaked.
 * @param entry entry to remove attribute from
 * @param type type of attribute to remove
 */
lsm_error lsm_entry_attr_remove(lsm_str **out, lsm_entry *entry, uint8_t type);

/****************************************
 *************** Handles ***************
 ***************************************/

struct lsm_read_handle {
  lsm_entry_wrapper *wrapper;
  lsm_store *store;

  struct {
    FILE *f;
    uint64_t pos;
  } data;
};

struct lsm_write_handle {
  lsm_entry_wrapper *wrapper;
  lsm_store *store;

  lsm_entry *dirty;
  bool removed;

  struct {
    FILE *f;
    uint64_t pos;
  } data;
};

/**
 * Allocate a new `lsm_read_handle` object
 */
lsm_error lsm_read_handle_init(lsm_read_handle **out);

/**
 * Allocate a new `lsm_write_handle` object
 */
lsm_error lsm_write_handle_init(lsm_write_handle **out);

/**
 * Commit changes solely to the memory part of the store.
 */
void lsm_write_commit_mem(lsm_write_handle *handle);

#endif
