#include <pthread.h>
#include <stdlib.h>
#include <string.h>

#include "lsm.h"
#include "lsm/store.h"
#include "lsm/store_internal.h"
#include "lsm/trie.h"

lsm_error lsm_store_init(lsm_store **ptr) {
  lsm_store *store = calloc(1, sizeof(lsm_store));

  if (store == NULL) {
    return lsm_error_failed_alloc;
  }

  LSM_RES2(lsm_trie_init(&store->trie), free(store));

  pthread_mutex_init(&store->db.lock, NULL);
  pthread_mutex_init(&store->idx.lock, NULL);

  *ptr = store;

  return lsm_error_ok;
}

uint64_t lsm_store_size(const lsm_store *store) {
  return lsm_trie_size(store->trie);
}

lsm_error lsm_store_open_read(lsm_read_handle **out, lsm_store *store,
                              const lsm_str *key) {
  lsm_entry_wrapper *wrapper;

  LSM_RES(lsm_trie_search((void **)&wrapper, store->trie, key));

  // Try to get a read lock on the entry's lock
  if (pthread_rwlock_tryrdlock(&wrapper->lock) != 0) {
    return lsm_error_lock_busy;
  }

  // While the trie's data field will never be NULL, the actual entry pointer
  // might be
  if (wrapper->entry == NULL) {
    pthread_rwlock_unlock(&wrapper->lock);

    return lsm_error_not_found;
  }

  lsm_read_handle *handle;
  LSM_RES2(lsm_read_handle_init(&handle),
           pthread_rwlock_unlock(&wrapper->lock));

  handle->wrapper = wrapper;
  handle->store = store;
  *out = handle;

  return lsm_error_ok;
}

lsm_error lsm_store_open_write(lsm_write_handle **out, lsm_store *store,
                               const lsm_str *key) {
  lsm_entry_wrapper *wrapper;
  LSM_RES(lsm_trie_search((void **)&wrapper, store->trie, key));

  // Try to get a write lock on the entry's lock
  // TODO make this timeout to not block
  if (pthread_rwlock_wrlock(&wrapper->lock) != 0) {
    return lsm_error_lock_busy;
  }

  // While the trie's data field will never be NULL, the actual entry pointer
  // might be
  if (wrapper->entry == NULL) {
    pthread_rwlock_unlock(&wrapper->lock);

    return lsm_error_not_found;
  }

  lsm_write_handle *handle;
  LSM_RES2(lsm_write_handle_init(&handle),
           pthread_rwlock_unlock(&wrapper->lock));

  handle->wrapper = wrapper;
  handle->store = store;
  *out = handle;

  return lsm_error_ok;
}

lsm_error lsm_store_open_new(lsm_write_handle **out, lsm_store *store,
                             lsm_str *key) {
  // TODO what happens when two inserts to the same key happen at the same time?
  lsm_entry_wrapper *wrapper;

  // If a key was previously removed from the trie, the wrapper will already be
  // present in the trie
  if (lsm_trie_search((void **)&wrapper, store->trie, key) ==
      lsm_error_not_found) {
    LSM_RES(lsm_entry_wrapper_init(&wrapper));

    pthread_rwlock_wrlock(&wrapper->lock);

    LSM_RES2(lsm_trie_insert(store->trie, key, wrapper),
             lsm_entry_wrapper_free(wrapper));
  } else {
    pthread_rwlock_wrlock(&wrapper->lock);

    if (wrapper->entry != NULL) {
      pthread_rwlock_unlock(&wrapper->lock);

      return lsm_error_already_present;
    }
  }

  lsm_entry *entry;
  LSM_RES2(lsm_entry_init(&entry), pthread_rwlock_unlock(&wrapper->lock));

  entry->key = key;

  lsm_write_handle *handle;
  LSM_RES2(lsm_write_handle_init(&handle),
           pthread_rwlock_unlock(&wrapper->lock));

  handle->wrapper = wrapper;
  handle->store = store;
  handle->dirty = entry;

  *out = handle;

  return lsm_error_ok;
}
