#include <stdint.h>

#include "lsm/store_internal.h"

static lsm_error lsm_fwrite(uint64_t *sum, FILE *f, uint64_t size,
                            uint64_t count, const void *val) {
  size_t res = fwrite(val, size, count, f);

  if (res < count) {
    return lsm_error_failed_io;
  }

  if (sum != NULL) {
    *sum += size * count;
  }

  return lsm_error_ok;
}

static lsm_error lsm_write_str(uint64_t *sum, FILE *f, const lsm_str *s) {
  uint64_t len = lsm_str_len(s);

  LSM_RES(lsm_fwrite(sum, f, sizeof(uint64_t), 1, &len));

  uint64_t written = 0;

  do {
    written += fwrite(lsm_str_ptr(s), sizeof(char), len - written, f);
  } while (written < len);

  if (sum != NULL) {
    *sum += len * sizeof(char);
  }

  return lsm_error_ok;
}

static lsm_error lsm_fseek(FILE *f, uint64_t pos) {
  if (fseek(f, pos, SEEK_SET) != 0) {
    return lsm_error_failed_io;
  }

  return lsm_error_ok;
}

lsm_error lsm_write_db_entry(uint64_t *size, FILE *db_file, lsm_entry *entry,
                             uint64_t pos) {
  *size = 0;

  LSM_RES(lsm_fseek(db_file, pos));

  LSM_RES(lsm_write_str(size, db_file, entry->key));
  LSM_RES(lsm_fwrite(size, db_file, sizeof(uint64_t), 1, &entry->data_len));
  LSM_RES(lsm_fwrite(size, db_file, sizeof(uint8_t), 1, &entry->attrs.count));

  for (uint8_t i = 0; i < entry->attrs.count; i++) {
    LSM_RES(lsm_fwrite(size, db_file, sizeof(uint8_t), 1,
                       &entry->attrs.items[i].type));
    LSM_RES(lsm_write_str(size, db_file, entry->attrs.items[i].str));
  }

  return lsm_error_ok;
}

lsm_error lsm_write_idx_entry(uint64_t *size, FILE *idx_file, uint64_t offset,
                              uint64_t len, uint64_t pos) {
  *size = 0;

  LSM_RES(lsm_fseek(idx_file, pos));

  LSM_RES(lsm_fwrite(size, idx_file, sizeof(uint64_t), 1, &offset));
  LSM_RES(lsm_fwrite(size, idx_file, sizeof(uint64_t), 1, &len));

  return lsm_error_ok;
}

lsm_error lsm_entry_disk_insert(lsm_store *store, lsm_entry *entry) {
  pthread_mutex_lock(&store->db.lock);

  uint64_t db_entry_index = store->db.size;

  uint64_t db_entry_size;
  lsm_error res =
      lsm_write_db_entry(&db_entry_size, store->db.f, entry, store->db.size);
  fflush(store->db.f);

  pthread_mutex_unlock(&store->db.lock);

  if (res != lsm_error_ok) {
    return res;
  }

  // Append entry to index file
  pthread_mutex_lock(&store->idx.lock);

  uint64_t idx_entry_index = store->idx.size;

  uint64_t idx_entry_size;
  res = lsm_write_idx_entry(&idx_entry_size, store->idx.f, db_entry_index,
                            db_entry_size, store->idx.size);

  if (res == lsm_error_ok) {
    // Update the counter at the beginning of the file
    rewind(store->idx.f);

    uint64_t new_block_count = store->idx.block_count + 1;

    res = lsm_fwrite(NULL, store->idx.f, sizeof(uint64_t), 1, &new_block_count);

    if (res == lsm_error_ok) {
      // Only if we successfully updated the on-disk counter do we make the code
      // aware that the files' sizes have increased. This way, if a write to the
      // counter fails, the code will simply reuse the already written content.
      store->idx.size += idx_entry_size;
      store->idx.block_count = new_block_count;
      store->db.size += db_entry_size;

      entry->idx_file_offset = idx_entry_index;
    }
  }

  fflush(store->idx.f);
  pthread_mutex_unlock(&store->idx.lock);

  return res;
}

static lsm_error lsm_idx_zero_block(lsm_store *store, uint64_t pos) {
  pthread_mutex_lock(&store->idx.lock);

  lsm_error res = lsm_fseek(store->idx.f, pos);

  if (res != lsm_error_ok) {
    pthread_mutex_unlock(&store->idx.lock);

    return res;
  }

  uint64_t val = 0;
  res = lsm_fwrite(NULL, store->idx.f, sizeof(uint64_t), 1, &val);

  pthread_mutex_unlock(&store->idx.lock);

  if (res != lsm_error_ok) {
    return res;
  }

  fflush(store->idx.f);

  return lsm_error_ok;
}

// Marking an entry as removed in the idx file is simply setting the length of
// its entry to zero
lsm_error lsm_entry_disk_remove(lsm_store *store, lsm_entry *entry) {
  LSM_RES(lsm_idx_zero_block(store, entry->idx_file_offset + sizeof(uint64_t)));
  LSM_RES(lsm_entry_data_remove(store, entry));

  return lsm_error_ok;
}

lsm_error lsm_entry_disk_update(lsm_store *store, lsm_entry *entry) {
  // An update is implemented by reinserting the entry at the end of the db file
  uint64_t old_idx_index = entry->idx_file_offset;

  // TODO is there any way we can make this atomic? If the zero write to the
  // index file fails, there are two entries in the db file for the same key.
  LSM_RES(lsm_entry_disk_insert(store, entry));
  LSM_RES(lsm_idx_zero_block(store, old_idx_index + sizeof(uint64_t)));

  return lsm_error_ok;
}
